/*
**  CXSC is a C++ library for eXtended Scientific Computing (V 2.1.1)
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2006 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* CVS $Id: l_imath.cpp,v 1.13 2006/03/08 14:16:27 cxsc Exp $ */

#include <math.h>
#include "l_imath.hpp"
#include "imath.hpp"
#include "rmath.hpp"

namespace cxsc {

#define  CXSC_One       1.
#define  CXSC_Zero      0.
#define  CXSC_MinusOne  -1.

l_interval pow(const l_interval & x, const l_interval & e) throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF)
{
   int         stagsave = stagprec,
               stagmax = 19,
               intexp;
               
   bool        fertig;

   l_interval  y;
   interval    dx = _interval(x),
               de = _interval(e),
               einfachgenau;
   real        supabsde = Sup(abs(de));

   einfachgenau = pow(dx,de);

   fertig = false;
   if (Inf(de) == Sup(de)) // &&
      if (supabsde < 32768.0) 
      {
         intexp = int(_double(_real(Sup(e))));
         if (_real(intexp) == Sup(e)) 
         {
            y = power(x,intexp);   // Integer-Potenz wesentlich schneller
            fertig = true;
         }
      }
      
   if (!fertig) 
   {
      if (Inf(x) < 0.0) 
         cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval pow(const l_interval & x, const l_interval & e)"));
      else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_One)
         y = x;
      else if (Inf(de) == Sup(de) && Sup(de) == CXSC_One)
         y = x;
      else if (Inf(de) == Sup(de) && Sup(de) == CXSC_Zero)
         y = 1.0;
      else 
      {
         if (stagprec < stagmax) 
            stagprec++;
         else                    
            stagprec = stagmax;
         y = exp(e*ln(x));
         stagprec = stagsave;
         y = adjust(y);
         y = y & einfachgenau;
      }
   }

   return y;
}

l_interval power(const l_interval &x, int n)       // Power(x,n)
{
   int         stagsave = stagprec,
               stagmax = 19;
              
   bool        neg = false;
              
   long int    zhi = 2;
   interval    dx = _interval(x),
               einfachgenau;
   l_interval  y, neu;

   einfachgenau = Power(dx,n);

   if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_One)
      y = x;
   else if (n == 0)
      y = adjust(_l_interval(1.0));
   else 
   {
      if (stagprec < stagmax) 
         stagprec++;
      else                    
         stagprec = stagmax;
      
      if (n == 1)
         y = x;
      else if (n == 2)
         y = sqr(x);
      else 
      {
         if (n < 0) 
         {
            neg = true;
            n = -n;
         }
         // Initialisierung
         if (n%2)  
            y = x;
         else      
            y = _l_interval(1.0);  // Praezision wird bei 1 Mult. auf
                                        // aktuellen Wert gesetzt;
         // Berechnugn durch binaere Darstellung der n
         neu = sqr(x);   // neu = x*x;
         do {
            if ((n/zhi)%2)  y *= neu;
            zhi += zhi;
            if (zhi <= n)  // letzte Mult. entfaellt --> schneller
               neu *= neu;
         } while (zhi <= n);

         if (neg) 
            y = 1.0/(y);
      }
      stagprec = stagsave;
      y = adjust(y);
      y = y & einfachgenau;
   }

   return y;
}

l_interval sqr(const l_interval & x)              // Sqr(x)
{
   l_interval	y;

   if (Inf(x) >= 0.0)       /* result = [x.inf*x.inf, x.sup*x.sup] */
      y = x * x;
   else if (Sup(x)<= 0.0) 
   { 
      /* result = [x.sup*x.sup, x.inf*x.inf] */
      y = _l_interval (-Sup(x) , -Inf(x));
      y = (y) * (y);
   } else 
   {                   
      /* result = [0.0, max(x.sup*x.sup,x.inf*x.inf)] */
      if (abs(Inf(x)) >= abs(Sup(x)))
         y = _l_interval(0.0, abs(Inf(x)));
      else
         y = _l_interval(0.0, abs(Sup(x)));
      y = (y) * (y);
   }
	return y;
}

l_interval sqrt(const l_interval & x) 
                         throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF) // Sqrt(x)
{  // Blomquist: scaling with 2^ex is necessary if expo(Sup(dx)) is too small!
   int         stagsave = stagprec,
               stagmax = 19,
               stagsave2;
   interval    dx = interval(x),
               einfachgenau;
   l_interval  a1,y,t,mt;
   bool Inf_Zero;  

   einfachgenau = sqrt(dx);

   if (Inf(x) < 0.0) 
      cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval sqrt(const l_interval & x)"));
   else if (Inf(dx) == Sup(dx) && 
                               (Sup(dx) == CXSC_Zero || Sup(dx) == CXSC_One))
      y = x;
   else 
   {   // scaling necessary if exponent ex < 0
       l_interval x1 = x;  
       Inf_Zero = (Inf(dx)==0);    
       if (Inf_Zero) x1 = Sup(x1); 
       int ex = expo(Sup(dx));
       if (ex>0) ex = 0;  // scaling unnecessary if ex>0
       else
       {
	   ex = -ex; // ex >= 0
	   if (ex > 1023) ex = 1023; // ex==1023 is sufficient
	   if (ex%2) ex--;  // ex>=0 is even
       }
       if (ex) times2pown(x1,ex);  // ex >= 0  --->  exact scaling!
      // Interval-Newton-methode: y = m(y)-f(m(y))/f'(y)
      t = sqrt(_interval(x1));
      if (stagprec < stagmax) 
         stagsave2 = stagprec+1;
      else                    
         stagsave2 = stagmax;
      stagprec = 1;
      while (stagprec < stagsave2) 
      { 
         stagprec += stagprec;
         if (stagprec > stagmax) 
            stagprec = stagmax;
         mt = mid(t);  times2pown(t,1);
         t = mt-((mt*mt-x1)/t);
      }
      if (ex) times2pown(t,-ex/2); // ex!=0 --> backscaling with 2^(-ex/2)
      stagprec = stagsave;   // restore the previous precision
      y = adjust(t);         // matching to previous stagprec.
      if (Inf_Zero) SetInf(y,0.0); 
      y = y & einfachgenau;  // seeking optimal inclusion with intersection
  }
  return y;
}

l_interval sqrt(const l_interval &x, int n) 
                throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF) // Sqrt(x,n)
// -1073741822 <= n <= +1073741823, sonst autom. Programm-Abbruch
// sqrt(x,n) jetzt mit Skalierung --> Hohe Genauigkeit in allen Bereichen!
// Blomquist, 28.12.03;
{  
    int stagsave = stagprec,
	stagshort, staglong, i, ex, N,
	stagmax = 19,
	max = 2;
    l_interval  my, fy, corr, y, xx;
    interval dx = _interval(x), einfachgenau;

    einfachgenau = sqrt(dx,n);

    if (Inf(x) < 0.0) 
	cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF
                        ("l_interval sqrt(const l_interval &x, int n)"));
    else if (stagprec == 1)
	y = pow(dx, _interval(1.0/n));
    else if (Inf(dx) == Sup(dx) && (Sup(dx) == CXSC_Zero || 
                                                   Sup(dx) == CXSC_One))
	y = x;
    else 
    {
	if (stagprec < stagmax) 
	    stagprec++;
	else stagprec = stagmax;
         
	while (max < stagprec) 
	    max += max;  // quadratische Konvergenz kann
                         // nicht eingehalten werden,
	max += max;      // deshalb eine Schleife mehr

	xx = x;
	ex = expo(Sup(dx));
	N = -ex;
        // Skalierung mit 2^N, so dass dx,xx etwa = 1; ----------------
	times2pown(dx,N);
	if (N>1023) { 
	    times2pown(xx,1023);
	    times2pown(xx,N-1023);
	}
	else times2pown(xx,N);
        // Skalierung beendet, Blomquist 28.12.03 --------------------- 

	y = pow(dx, _interval(1.0/n));
	stagprec = 1;
	// Intervall-Newton-Verfahren
	for (i = 2; i <= max; i += i) 
	{
	    // Verdoppelung der Genauigkeit:
	    stagshort = stagprec;
	    stagprec += stagprec; // Verdoppelung hier
	    if (stagprec > stagmax) 
		stagprec = stagmax;
	    staglong  = stagprec;
	    my = _l_interval(mid(y));
	    fy = power(my, n)-xx;
	    // Fehlerauswertung nur in halber Genauigkeit notwendig!
	    stagprec  = stagshort;
	    corr = fy/(_real(n)*power(y, n-1));
	    stagprec  = staglong;
	    // Fehlerkorrektur in normaler Genauigkeit:
	    y = my-corr;
	}
        // Rueckskalierung mit dem Faktor 2^(-N/n):
	fy = l_interval(-N)/n;
	y *= exp(fy*li_ln2());  // li_ln2() = ln(2)
	    
	stagprec = stagsave;
	y = adjust(y);  // Anpassung an Ausgangs-stagprec = stagsave.
	y = y & einfachgenau;  // Falls y breiter ist als einfachgenau
    }

    return y;
} // sqrt(x,n)

l_interval sqrt1px2(const l_interval& x) throw()
// Calculation of an optimal inclusion of sqrt(1+x^2); Blomquist, 13.12.02;
// With stagmax=19 we get about 16*19=304 exact decimal digits.
{   // First step: Calculation of sqrt(1+x*x) in simple type interval:
    interval einfach = sqrt1px2( interval(x) );  // Only interval types
    int stagsave, stagmax=19;
    stagsave = stagprec;
    if (stagprec > stagmax) stagprec = stagmax;
    // Second step: Inclusion of sqrt(1+x^2) with stagprec <= 19
    const int exmax=512;
    l_interval y = abs(x);
    interval z = interval(Sup(x));
    int ex = expo(Sup(z));
    if (ex > exmax) 
    {   // scaling to avoid overflow by sqr(y):
	ex = exmax - ex;    // ex = 512 - ex;
	times2pown(y,ex);   // scaling to avoid overflow by sqr(y)
	y = sqrt( comp(0.5,2*ex+1) + sqr(y) ); // sqr(y) without overflow!
	times2pown(y,-ex);  // backscaling of the result with 2^(-ex)
    } else
    { // no scaling necessary:
        y = sqrt(1+sqr(x));
    }
    stagprec = stagsave;  // restore the old stagprec value
    y = adjust(y);       // y gets the previous precision
    y = einfach & y; // This intersection delivers for intervals x with large
                     // diameters the optimal result inclusion
    return y;
} // sqrt1px2(...)

l_interval sqrtx2y2(const l_interval& x, const l_interval& y) throw()
// Inclusion of sqrt(x^2+y^2); Blomquist, 14.12.02;
{
    interval ia = abs(interval(x)), ib = abs(interval(y)), einfach;
    einfach = sqrtx2y2(ia,ib); // Inclusion only with type interval.
    if (!einfach) return l_interval(0.0);

    int stagsave=stagprec, stagmax=19;
    if (stagprec>stagmax) stagprec = stagmax;

    l_interval a=abs(x), b=abs(y), r;
    int exa=expo(Sup(ia)), exb=expo(Sup(ib)), ex;
    if (exb > exa)
    {  // Permutation of a,b:
	r = a;  a = b;  b = r;
	ex = exa;  exa = exb;  exb = ex;
    }
    ex = 511 - exa;   exa = 0;
    if (ex>1022)
    { exa = ex - 1022;  ex = 1022; }  // ex > 1022 --> scaling in two steps
    times2pown(a,ex);                 // is necessary!
    if (exa) times2pown(a,exa);
    times2pown(b,ex);            // First step:  scaling b with 2^ex;
    if (exa) times2pown(b,exa);  // Second step: scaling b with 2^exa;
    r = sqrt(a*a + b*b);
    times2pown(r,-ex);           // Backscaling, first step
    if (exa) times2pown(r,-exa); // Backscaling, second step
    stagprec = stagsave;
    r = adjust(r);
    r = einfach & r;
    return r;
} // sqrtx2y2

l_interval sqrtp1m1(const l_interval& x) throw(STD_FKT_OUT_OF_DEF)
// sqrtp1m1(x) calculates an inclusion of sqrt(x+1)-1;
// Blomquist, 05.08.03;
{
    int stagsave=stagprec, stagmax=19;
    stagprec++;
    if (stagprec>stagmax) stagprec = stagmax;
    l_interval y,tmp;
    interval z = interval(x); // z is an inclusion of x
    real r = Inf(z);
    if (r < -1)
    cxscthrow(STD_FKT_OUT_OF_DEF("l_interval sqrtp1m1(const l_interval&)"));
    const real c = 1e-10;
    tmp = x+1;
    y = x<=interval(-c,c) ? x / (sqrt(tmp)+1) : sqrt(tmp)-1;
    stagprec = stagsave;
    y = adjust(y);
    return y; 
} // sqrtp1m1


l_interval sin(const l_interval & x) throw(ERROR_LINTERVAL_FAK_OVERFLOW)    // Sin(x)
{
   int         stagsave = stagprec,
               stagmax = 19;
   l_interval  pihalbe,
               y;
   interval    dx = _interval(x),
               einfachgenau;

   einfachgenau = sin(dx);

   if (stagprec == 1) 
      y = sin(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
      y = adjust(_l_interval(0.0));
   else 
   {
      if (stagprec < stagmax) 
         stagprec++;
      else                    
         stagprec = stagmax;
      
//      pihalbe = 2.0*atan(_l_interval(1.0));
      pihalbe = li_pi4();
      times2pown(pihalbe,1); // Blomquist, 05.12.03;
      y = x-pihalbe;
      
      try {
         y = cos(y);
      }
      catch(const ERROR_LINTERVAL_FAK_OVERFLOW &) // Damit Funktionsname stimmt
      {
         cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval sin(const l_interval & x)")); 
      }
      
      
      stagprec = stagsave;
      y = adjust(y);
      y = y & einfachgenau;
   }

   return y;
}

l_interval cos(const l_interval & x) throw(ERROR_LINTERVAL_FAK_OVERFLOW)   // Cos(x)
{
   long int    mm = 6;
   int         stagsave = stagprec,
               stagmax = 19,
               n = 0,
               digits = 53,
               sign = 0,
               degree, k,
               m2;
               
   bool        xinf = false,
               xsup = false,
               fertig = false;

   real        abst, m, eps,
//               bas, tn, t4,
               lneps, lnt,
               zk,
               zhn = 1.0,
               lnb = 0.69314718,
               fak = 720.0;    // 6!
   interval    dx = _interval(x),
               einfachgenau,
               dt, extr, error;
   l_interval  zwopi, t, t2, p, ph, test,
               y;

   einfachgenau = cos(dx);
   if (stagprec == 1) 
      y = cos(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
      y = adjust(_l_interval(1.0));
   else 
   {
      if (stagprec < stagmax) 
         stagprec++;
      else                    
         stagprec = stagmax;

      // stagprec++;
      // zwopi = 8.0*atan(_l_interval(1.0));
      zwopi = li_pi4();
      times2pown(zwopi,3); // Blomquist, 05.12.03;
      // stagprec--;

      // erste Argumentreduktion ( cos(x) = cos(x+2k*pi) )
      if (Sup(zwopi) < Sup(abs(x))) 
      {
         m = floor(_double(_real(Sup((x/zwopi)))));  // floor rundet zur nchsten
         t = x-m*zwopi;                              // ganzen Zahl kleiner x
         if (Sup(zwopi) < Sup(abs(t))) 
         { // das ganze nochmal, falls m uebergelaufen ist!
                                      // ohne Sup wird Inclusion geprueft!
            m = floor(_double(_real(Sup((t/zwopi)))));// rundet zur naechsten Zahl < x
            t = t-m*zwopi;
         }
      } else 
         t = x;

      // ueberpruefen, ob Maximum oder Minimum im Inneren des Intervalls
      extr = _interval(2.0/zwopi*t);
      m2 = int(_double(floor(_double(Sup(extr)))));
      if (_interval(_real(m2)) <= extr) 
      {
         if (_interval(_real(m2-1),_real(m2)) <= extr) 
         {
            y = _l_interval(-1.0,1.0);
            fertig = true;
         } else 
            if (m2%2)  
               xinf = TRUE;
            else            
               xsup = TRUE;
      }

      if (!(fertig)) 
      {
         // zweite Argumentreduktion
         dt = _interval(t);
         // eps = 0.01/(stagprec*stagprec*stagprec);
         eps = 0.01;
         while (Sup(abs(dt))/zhn >= eps) 
         {
            n++;
            zhn += zhn;
         }
         t /= zhn;

         // Abschaetzung der Genauigkeit

         t2 = t*t;
         abst = _real(Sup(abs(t)));
         if (abst < MinReal) 
            abst = MinReal; // nur zur Sicherheit
         lnt = ln(abst);
         lneps = (1.0-digits*stagprec)*lnb;
         while (lneps-(_real(mm)*lnt+ln(2.0/fak)) <= 0.0) 
         {
            mm  += 4;
            if (mm > 170) 
            {   
               // 170! = 7.26*10^306
               cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval cos(const l_interval & x)"));
               mm = 170;
               break;
            }
            fak *= mm*(mm-1.0)*(mm-2.0)*(mm-3.0);
         }
         /*
            bas = _real(Inf(power(_l_interval(2.0),(1-digits*stagprec))));
            tn = abs(_real(Sup(power(t,6))));
            t4 = _real(mid(power(t,4)));
            while (bas-2.0*tn/fak <= 0.0) 
            {
               mm += 4;
               if (mm > 170) 
               {    // 170! = 7.26*10^306
                  errmon (ERROR_LINTERVAL(FAKOVERFLOW));
                  mm = 170;
                  break;
            }
            tn *= t4;
            fak *= mm*(mm-1.0)*(mm-2.0)*(mm-3.0);
         }
         */

         degree = mm-2;     // Achtung mm := 2n+2 !

         // Polynomauswertung

         sign = (degree/2)%2;
         zk  = _real(degree)*_real(degree-1);
         if (sign) 
            p = -t2/zk;
         else      
            p = t2/zk;
         for (k = degree-2; k >= 2; k -= 2) 
         {
            sign = 1-sign;
            if (sign)   
               p -= 1.0;
            else       
               p += 1.0;
            zk = _real(k)*_real(k-1);
            p *= t2/zk;
         }

         error = pow(_interval(2.0), _interval(1.0-digits*stagprec))
                 * _interval(-0.5,0.5);
         p = _l_interval(1.0)+error+p;

         // Rckgngigmachung der zweiten Argumentreduktion
         for (int i = 0; i < n; i++) 
            p = 2.0*p*p-1.0;

         stagprec = stagsave;
         y = adjust(p);
         if (Inf(y) < -1.0) 
            SetInf(y,-1.0);
         if (Sup(y) > 1.0) 
            SetSup(y,1.0);
         if (xinf) 
            SetInf(y,-1.0);
         if (xsup) 
            SetSup(y,1.0);
      }
      y = y & einfachgenau;
   }
   return y;
}

l_interval tan(const l_interval & x) throw(ERROR_LINTERVAL_FAK_OVERFLOW,ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF)   // Tan(x)
{
   interval    dx = _interval(x),
               einfachgenau;
   l_interval  s, c, y;

   einfachgenau = tan(dx);

   if (stagprec == 1) 
      y = tan(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
         y = adjust(_l_interval(0.0));
   else 
   {
      try 
      {
         c = cos(x);
      }
      catch(const ERROR_LINTERVAL_FAK_OVERFLOW &) // Damit Funktionsname stimmt
      {
         cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval tan(const l_interval &x)"));
      }

      if (_interval(0.0) <= c) 
      {
         cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval tan(const l_interval &x)"));
      }
      try
      {
         s = sin(x);
      }
      catch(const ERROR_LINTERVAL_FAK_OVERFLOW &)
      {
         cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval tan(const l_interval &x)"));
      }
      stagprec++;
      y = s/c;
      stagprec--;
      y = adjust(y);
      y = y & einfachgenau;
   }
   return y;
}

l_interval cot(const l_interval & x) throw(ERROR_LINTERVAL_FAK_OVERFLOW,ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF)   // Cot(x)
{
   interval    dx = _interval(x),
               einfachgenau;
   l_interval  s, c, y;

   einfachgenau = cot(dx);

   if (stagprec == 1) 
      y = tan(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
      y = adjust(_l_interval(0.0));
   else 
   {
      try
      {
         s = sin(x);
      }
      catch(const ERROR_LINTERVAL_FAK_OVERFLOW &) // Damit Funktionsname stimmt
      {
         cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval cot(const l_interval &x)"));
      }

      if (_interval(0.0) <= s) 
      {
         cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval cot(const l_interval &x)"));
      }

      try 
      {
         c = cos(x);
      }
      catch(const ERROR_LINTERVAL_FAK_OVERFLOW &) // Damit Funktionsname stimmt
      {
         cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval cot(const l_interval &x)"));
      }
   
      c = cos(x);
      stagprec++;
      y = c/s;
      stagprec--;
      y = adjust(y);
      y = y & einfachgenau;
   }

   return y;
}

l_interval asin(const l_interval & x) throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF)  // ASin(x)
{
   l_interval  t, ta, u, pihalbe,
               y;
   interval    dx = _interval(x),
               einfachgenau;
   real        supabsdx = Sup(abs(dx)),
               infdx = Inf(dx),
               supdx = Sup(dx);

   einfachgenau = asin(dx);

   stagprec++;
//   pihalbe = 2.0*atan(_l_interval(1.0));
   pihalbe = li_pi4();
   times2pown(pihalbe,1); // Blomquist, 05.12.03;
   stagprec--;
   
   if (Inf(x) < CXSC_MinusOne || Sup(x) > CXSC_One) 
      cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval asin(const l_interval & x)"));
   else if (stagprec == 1) 
      y = asin(dx);
   else if (infdx == supdx && supdx == CXSC_Zero)
      y = adjust(_l_interval(0.0));
   else if (infdx == supdx && supabsdx == CXSC_One) 
   {
      if (supdx == 1.0) 
         y = pihalbe;
      else              
         y = -pihalbe;
   } else 
   {
      stagprec++;
      try 
      {
         if (supabsdx <= 0.75) 
            u = x;
         else                  
            u = 2.0*x*sqrt((1.0-x)*(1.0+x));
         t = u/sqrt((1.0-u)*(1.0+u));
         stagprec--;
         ta = atan(t);
      }
      catch(const ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF &)
      {
         cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval asin(const l_interval & x)"));
      }
      stagprec++;
      if (supabsdx <= 0.75) 
         y = ta;
      else if (Sup(t) >= 0.0)  
         y = pihalbe-0.5*ta;
      else 
         y = -pihalbe-0.5*ta;
      stagprec--;
      y = adjust(y);
      y = y & einfachgenau;
   }

   return y;
}

l_interval acos(const l_interval & x) throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF)   // ACos(x)
{
   bool        neg=false;
   l_interval  pi, s, y;
   interval    dx = _interval(x),
               einfachgenau;
   real        supabsdx = Sup(abs(dx)),
               infdx = Inf(dx),
               supdx = Sup(dx);

   try {

      einfachgenau = acos(dx);

      stagprec++;
//      pi = 4.0*atan(_l_interval(1.0));
      pi = li_pi4();
      times2pown(pi,2); // Blomquist, 05.12.03;
      stagprec--;
      if (Inf(x) < CXSC_MinusOne || Sup(x) > CXSC_One) 
         cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF());
      else if (stagprec == 1) 
         y = acos(dx);
      else if (infdx == supdx && supabsdx == CXSC_One) 
      {
         if (supdx == 1.0) 
            y = adjust(_l_interval(0.0));
         else              
            y = adjust(pi);
      } else 
      {
         if (supdx < 0.0) 
         {
            y = -x;
            neg = true;
         } else 
         {
            y = x;
            neg = false;
         }
         stagprec++;
         if (supabsdx > 0.75) 
         {
            y = (1.0-(y))*(1.0+(y));
            //      stagprec--;
            y = asin(sqrt(y));
            //      stagprec++;
            if (neg)  
               y = pi-(y);
         } else 
         {
            //      stagprec--;
            s = asin(y);
            //      stagprec++;
            if (neg)  
               y = 0.5*pi+s;
            else      
               y = 0.5*pi-s;
         }

         // Fehler in der Systemumgebung, deshalb die Aufblhung des Intervalls
         real      err1, err2;
         interval  error;
         err1 = 5.0*_real(diam(y));
         err2 = 5.0*power(10.0,-16*(stagprec-1)-1);
         if (err1 > err2)  
            error = _interval(-err1, err1);
         else              
            error = _interval(-err2, err2);
         y += error;

         stagprec--;
         y = adjust(y);
         y = y & einfachgenau;
      }
   }
   catch(const ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF &)
   {
      cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval acos(const l_interval & x)"));
   }
   return y;
}

static real CXSC_ln2[21]; // CXSC_ln2[0], ... CXSC_ln2[20] 
static bool CXSC_ln2_initialized = false;

l_interval li_ln2() throw()
// Inclusion of ln(2), Blomquist, 04.12.03;
{
   l_interval y;
   int stagsave = stagprec,
       stagmax = 20;
   
   if (!CXSC_ln2_initialized) 
   {
       std::string str;
       std::cout << SaveOpt;
       std::cout << Hex;
       str = "+162E42FEFA39EFe3FE";
       str >> CXSC_ln2[0];
       str = "+1ABC9E3B39803Fe3C7";
       str >> CXSC_ln2[1];
       str = "+17B57A079A1934e390";
       str >> CXSC_ln2[2];
       str = "-1ACE93A4EBE5D1e35A";
       str >> CXSC_ln2[3];
       str = "-123A2A82EA0C24e324";
       str >> CXSC_ln2[4];
       str = "+1D881B7AEB2615e2ED";
       str >> CXSC_ln2[5];
       str = "+19552FB4AFA1B1e2B7";
       str >> CXSC_ln2[6];
       str = "+1DA5D5C6B82704e27E";
       str >> CXSC_ln2[7];
       str = "+14427573B29117e247";
       str >> CXSC_ln2[8];
       str = "-191F6B05A4D7A7e211";
       str >> CXSC_ln2[9];
       str = "-1DB5173AE53426e1DB";
       str >> CXSC_ln2[10];
       str = "+11317C387EB9EBe1A3";
       str >> CXSC_ln2[11];
       str = "-190F13B267F137e16D";
       str >> CXSC_ln2[12];
       str = "+16FA0EC7657F75e137";
       str >> CXSC_ln2[13];
       str = "-1234C5E1398A6Be101";
       str >> CXSC_ln2[14];
       str = "+1195EBBF4D7A70e0CA";
       str >> CXSC_ln2[15];
       str = "+18192432AFD0C4e094";
       str >> CXSC_ln2[16];
       str = "-1A1BE38BA4BA4De05E";
       str >> CXSC_ln2[17];
       str = "-1D7860151CFC06e024";
       str >> CXSC_ln2[18];
       str = "+1000032847ED6Fe000";
       str >> CXSC_ln2[19];
       str = "+1000032847ED70e000";
       str >> CXSC_ln2[20];

       CXSC_ln2_initialized = true;
       std::cout << RestoreOpt;
   }
   stagprec = stagmax;
   y = adjust(l_interval(0.0)); 
   for (int i=0; i <= stagmax; i++)  
       y.data[i] = CXSC_ln2[i];
   stagprec = stagsave;
   y = adjust(y);
   return y;             
}

static real CXSC_ln10[21]; // CXSC_ln10[0], ... CXSC_ln10[20] 
static bool CXSC_ln10_initialized = false;

l_interval li_ln10() throw()
// Inclusion of ln(2), Blomquist, 04.12.03;
{
   l_interval y;
   int stagsave = stagprec,
       stagmax = 20;
   
   if (!CXSC_ln10_initialized) 
   {
       std::string str;
       std::cout << SaveOpt;
       std::cout << Hex;
       str = "+126BB1BBB55516e400";
       str >> CXSC_ln10[0];
       str = "-1F48AD494EA3E9e3CA";
       str >> CXSC_ln10[1];
       str = "-19EBAE3AE0260Ce394";
       str >> CXSC_ln10[2];
       str = "-12D10378BE1CF1e35E";
       str >> CXSC_ln10[3];
       str = "+10403E05AE52C6e328";
       str >> CXSC_ln10[4];
       str = "-1FA509CAFDF466e2F0";
       str >> CXSC_ln10[5];
       str = "-1C79A1FE9D0795e2BA";
       str >> CXSC_ln10[6];
       str = "+1058C448308218e284";
       str >> CXSC_ln10[7];
       str = "-1D250470877BFDe24D";
       str >> CXSC_ln10[8];
       str = "-1AE92987D3075De215";
       str >> CXSC_ln10[9];
       str = "-1D5CDBB8626956e1DF";
       str >> CXSC_ln10[10];
       str = "-13C4F27CE0410Ae1A9";
       str >> CXSC_ln10[11];
       str = "+1B3AC12ACF1BE9e173";
       str >> CXSC_ln10[12];
       str = "+1161BB49D219C8e13D";
       str >> CXSC_ln10[13];
       str = "-110D6613293728e107";
       str >> CXSC_ln10[14];
       str = "+142163A4CDA351e0CF";
       str >> CXSC_ln10[15];
       str = "+1E2713D6C22C16e097";
       str >> CXSC_ln10[16];
       str = "-15090EF85CB0ADe05E";
       str >> CXSC_ln10[17];
       str = "-1C5B3E859F876Ee027";
       str >> CXSC_ln10[18];
       str = "-10000703552C52e000";
       str >> CXSC_ln10[19];
       str = "-10000703552C51e000";
       str >> CXSC_ln10[20];

       CXSC_ln10_initialized = true;
       std::cout << RestoreOpt;
   }
   stagprec = stagmax;
   y = adjust(l_interval(0.0)); 
   for (int i=0; i <= stagmax; i++)  
       y.data[i] = CXSC_ln10[i];
   stagprec = stagsave;
   y = adjust(y);
   return y;             
}

static real CXSC_Rln10[21]; // CXSC_Rln10[0], ... CXSC_Rln10[20] 
static bool CXSC_Rln10_initialized = false;

l_interval li_Rln10() throw()
// Inclusion of ln(2), Blomquist, 04.12.03;
{
   l_interval y;
   int stagsave = stagprec,
       stagmax = 20;
   
   if (!CXSC_Rln10_initialized) 
   {
       std::string str;
       std::cout << SaveOpt;
       std::cout << Hex;
       str = "+1BCB7B1526E50Ee3FD";
       str >> CXSC_Rln10[0];
       str = "+195355BAAAFAD3e3C6";
       str >> CXSC_Rln10[1];
       str = "+1EE191F71A3012e38F";
       str >> CXSC_Rln10[2];
       str = "+17268808E8FCB5e358";
       str >> CXSC_Rln10[3];
       str = "+13DE3A94F1D509e320";
       str >> CXSC_Rln10[4];
       str = "+1DF42805E7E524e2E9";
       str >> CXSC_Rln10[5];
       str = "+11AAC96323250Be2B3";
       str >> CXSC_Rln10[6];
       str = "-1CE63884C058E4e27D";
       str >> CXSC_Rln10[7];
       str = "-1A1C82EA3969BAe247";
       str >> CXSC_Rln10[8];
       str = "+1B4F6686AD7A33e211";
       str >> CXSC_Rln10[9];
       str = "-1B97C8035FFC70e1DB";
       str >> CXSC_Rln10[10];
       str = "+1630771369962Ee1A0";
       str >> CXSC_Rln10[11];
       str = "-1E15BD37B295AFe16A";
       str >> CXSC_Rln10[12];
       str = "-132484B432318Be134";
       str >> CXSC_Rln10[13];
       str = "+15430212AE68C0e0FE";
       str >> CXSC_Rln10[14];
       str = "+1351923B322731e0C8";
       str >> CXSC_Rln10[15];
       str = "+11F934D794D64Fe092";
       str >> CXSC_Rln10[16];
       str = "+13E4B475D9FF20e05B";
       str >> CXSC_Rln10[17];
       str = "+185D9B63ED9A24e025";
       str >> CXSC_Rln10[18];
       str = "+1000035B8CA18Ce000";
       str >> CXSC_Rln10[19];
       str = "+1000035B8CA18De000";
       str >> CXSC_Rln10[20];

       CXSC_Rln10_initialized = true;
       std::cout << RestoreOpt;
   }
   stagprec = stagmax;
   y = adjust(l_interval(0.0)); 
   for (int i=0; i <= stagmax; i++)  
       y.data[i] = CXSC_Rln10[i];
   stagprec = stagsave;
   y = adjust(y);
   return y;             
}

static real    CXSC_pi4[21];
static bool    CXSC_pi4_initialized = false;

l_interval li_pi4() throw()
// Inclusion of ln(2), Blomquist, 04.12.03;
{
   l_interval y;
   int stagsave = stagprec,
       stagmax = 20;
   
   if (!CXSC_pi4_initialized) 
   {
       std::string str;
       std::cout << SaveOpt;
       std::cout << Hex;
       str = "+1921FB54442D18e3FE";
       str >> CXSC_pi4[0];
       str = "+11A62633145C06e3C8";
       str >> CXSC_pi4[1];
       str = "+1C1CD129024E08e393";
       str >> CXSC_pi4[2];
       str = "+114CF98E804178e35E";
       str >> CXSC_pi4[3];
       str = "-159C4EC64DDAECe327";
       str >> CXSC_pi4[4];
       str = "+1410F31C6809BCe2F2";
       str >> CXSC_pi4[5];
       str = "-106AE64C32C5BCe2BB";
       str >> CXSC_pi4[6];
       str = "-1C99FA9EB241B4e286";
       str >> CXSC_pi4[7];
       str = "-1D791603D95252e24E";
       str >> CXSC_pi4[8];
       str = "-1571EDD0DBD254e218";
       str >> CXSC_pi4[9];
       str = "-133B4302721768e1E2";
       str >> CXSC_pi4[10];
       str = "+10BA698DFB5AC2e1AD";
       str >> CXSC_pi4[11];
       str = "+1FFAE5B7A035C0e178";
       str >> CXSC_pi4[12];
       str = "-1211C79404A576e143";
       str >> CXSC_pi4[13];
       str = "-1816945836FBA0e10D";
       str >> CXSC_pi4[14];
       str = "-1DA700CDB6BCCEe0D8";
       str >> CXSC_pi4[15];
       str = "+11ECE45B3DC200e0A3";
       str >> CXSC_pi4[16];
       str = "+1F2E2858EFC166e06D";
       str >> CXSC_pi4[17];
       str = "+1B4906C38ABA73e036";
       str >> CXSC_pi4[18];
       str = "+19A458FEA3F493e000";
       str >> CXSC_pi4[19];
       str = "+19A458FEA3F494e000";
       str >> CXSC_pi4[20];

       CXSC_pi4_initialized = true;
       std::cout << RestoreOpt;
   }
   stagprec = stagmax;
   y = adjust(l_interval(0.0)); 
   for (int i=0; i <= stagmax; i++)  
       y.data[i] = CXSC_pi4[i];
   stagprec = stagsave;
   y = adjust(y);
   return y;             
}

static real    CXSC_sqrt2[21];
static bool    CXSC_sqrt2_initialized = false;

l_interval li_sqrt2() throw()
// Inclusion of ln(2), Blomquist, 04.12.03;
{
   l_interval y;
   int stagsave = stagprec,
       stagmax = 20;
   
   if (!CXSC_sqrt2_initialized) 
   {
       std::string str;
       std::cout << SaveOpt;
       std::cout << Hex;
       str = "+16A09E667F3BCDe3FF";
       str >> CXSC_sqrt2[0];
       str = "-1BDD3413B26456e3C9";
       str >> CXSC_sqrt2[1];
       str = "+157D3E3ADEC175e393";
       str >> CXSC_sqrt2[2];
       str = "+12775099DA2F59e35B";
       str >> CXSC_sqrt2[3];
       str = "+160CCE64552BF2e322";
       str >> CXSC_sqrt2[4];
       str = "+1821D5C5161D46e2E9";
       str >> CXSC_sqrt2[5];
       str = "-1C032046F8498Ee2B3";
       str >> CXSC_sqrt2[6];
       str = "+1EE950BC8738F7e27B";
       str >> CXSC_sqrt2[7];
       str = "-1AC3FDBC64E103e245";
       str >> CXSC_sqrt2[8];
       str = "+13B469101743A1e20D";
       str >> CXSC_sqrt2[9];
       str = "+15E3E9CA60B38Ce1D7";
       str >> CXSC_sqrt2[10];
       str = "+11BC337BCAB1BDe19C";
       str >> CXSC_sqrt2[11];
       str = "-1BBA5DEE9D6E7De166";
       str >> CXSC_sqrt2[12];
       str = "-1438DD083B1CC4e130";
       str >> CXSC_sqrt2[13];
       str = "+1B56A28E2EDFA7e0FA";
       str >> CXSC_sqrt2[14];
       str = "+1CCB2A634331F4e0C4";
       str >> CXSC_sqrt2[15];
       str = "-1BD9056876F83Ee08D";
       str >> CXSC_sqrt2[16];
       str = "-1234FA22AB6BEFe057";
       str >> CXSC_sqrt2[17];
       str = "+19040CA4A81395e020";
       str >> CXSC_sqrt2[18];
       str = "-1000002A493818e000";
       str >> CXSC_sqrt2[19];
       str = "-1000002A493817e000";
       str >> CXSC_sqrt2[20];

       CXSC_sqrt2_initialized = true;
       std::cout << RestoreOpt;
   }
   stagprec = stagmax;
   y = adjust(l_interval(0.0)); 
   for (int i=0; i <= stagmax; i++)  
       y.data[i] = CXSC_sqrt2[i];
   stagprec = stagsave;
   y = adjust(y);
   return y;             
}

l_interval atan(const l_interval & x) throw()  // aTan(x)
{
   int         stagsave = stagprec,
               stagmax = 19,
               digits = 53,
               m = 1,
               stagsave2,
               degree, sign, zk;
   real        lnt2, lneps, ln3m, eps,
   //            bas, tn, t3,
               zhn = 1.0,
               lnb = 0.69314718, // ln(2.0)
               ln3 = 1.098612289;  // ln(3.0), Blomquist 27.12.03
   l_interval  t, t2, p,
               y;
   interval    dx = _interval(x),
               einfachgenau,
               err;

   einfachgenau = atan(dx);

   if (stagprec == 1) 
      y = atan(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
      y = adjust(_l_interval(0.0));
   else 
   {
       if (dx == _interval(1.0)) y = li_pi4(); // y = Pi/4
       else 
       {
	   if (stagprec < stagmax) 
	       stagprec++;
	   else                    
	       stagprec = stagmax;

           // Argumentreduktion
	   t = x;
	   eps = 0.01/stagprec;    // vorher 0.1, aber zu gross
	   while (Sup(abs(_interval(t))) > eps) 
	   {  // t = t/(1.0+sqrt(1.0+t*t));
	       t = t/(1.0+sqrt1px2(t)); // Blomquist, 13.12.02;
	       zhn += zhn;
	   }
	   t2 = t*t;

           // Bestimmung des Approximationsgrades
           // Beginn Blomquist, 27.12.03 ------------------------------
	   err = interval(abs(t2));
	   if (expo(Sup(err)) < -300) m = 4; // Vermeidet alte Fehlermeldg.
	   else 
	   {   
	       lnt2 = Sup(ln(err)); // Erzeugt alten Fehler nicht mehr!! 
	       ln3m = ln(3.0-_real(Sup(t2)));
	       lneps = (1.0-digits*stagprec)*lnb;
	       do {
		   m += 3;
	       } while (lneps-ln3-_real(m+1)*lnt2+ln(2.0*m+3.0)+ln3m <= 0.0);
	   }
           // Ende Blomquist, 27.12.03 ---------------------------------
         /*
            lnt = Sup(ln(_interval(abs(t)))); // bei Andreas u. Christian "t2"
            lneps = (1.0-digits*stagprec)*lnb;
            do {
               m += 3;
            } while (lneps+ln(2*m+3)-lnt*(2*m+3) <= 0.0);

            bas = _real(power(_l_real(2.0),1-digits*stagprec));
            tn = abs(_real(Sup(t)));
            t3 = abs(_real(Sup(power(t,3))));
            do {
               m += 3;
               tn *= t3;
               if (tn == 0.0) 
                  break; // bei Unterlauf erfolgt keine Fehlermeldung!
            } while (bas-tn/m <= 0.0);
         */
	   degree = m;

	   // Approximation
	   sign = (degree%2) ? -1 : 1;
	   zk = sign*(2*degree+1);
	   p = _l_interval(1.0)/_real(zk);
	   for (int k = degree-1; k >= 0; k--) 
	   {
	       sign = -sign;
	       zk = sign*(2*k+1);
	       p = p*t2+_l_interval(1.0)/_real(zk);
	   } // letztes t wird beim Fehler mit einmultipliziert

	   // Fehler bestimmen
	   stagsave2 = stagprec;
	   stagprec = 2;
	   l_interval relerr;
	   stagprec = stagsave2;
	   err = pow(_interval(2.0), _interval(1.0-digits*stagprec))
	       * _interval(-1.0, 1.0);      // von 2^(2-d*s) ge\x{201E}ndert !
	   relerr[1] = 1.0;
	   relerr[2] = Inf(err);
	   relerr[3] = Sup(err);

           // Argumentreduktion rckgngig machen, mit t mult. 
           // und Fehler einbauen
	   y = zhn*t*p*relerr;

	   stagprec = stagsave;
	   y = adjust(y);
	   y = y & einfachgenau;
       }
   }

   return y;
}

l_interval acot(const l_interval &x) throw()  // ACot(x)
{
   l_interval  pihalbe, y;
   interval    dx = _interval(x),
               einfachgenau;

   einfachgenau = acot(dx);
   stagprec++;
//   pihalbe = 2.0*atan(_l_interval(1.0));
   pihalbe = li_pi4();
   times2pown(pihalbe,1); // Blomquist, 05.12.03;
   stagprec--;

   if (stagprec == 1)
      y = acot(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
      y = adjust(pihalbe);
   else 
   {
      y = pihalbe-atan(x);
      // y = atan(1.0/x);  andere Mglichkeit der Berechnung
      y = adjust(y);
      y = y & einfachgenau;
   }

   return y;
}

l_interval exp(const l_interval & x) throw(ERROR_LINTERVAL_FAK_OVERFLOW) // exp(x)
{
   long int    n = 2;
   int         stagsave = stagprec,
               stagmax = 19,
               degree,
               rednum = 0,
               digits = 53;
   real        fak = 2.0,
               zhn = 1.0,       // kann groesser 32768 werden
               tmp, lny, lneps, eps,
               lnb = 0.69314718;
   l_interval  p, t,
               y;
   interval    dx = _interval(x),
               einfachgenau,
               dt, error;
   //            ilnb =_interval(0.6931471, 0.6931572),
   //            ilny, ilneps;

   einfachgenau = exp(dx);

   // gueltigen Bereich pruefen
   if (stagprec == 1) 
      y = exp(dx);
   else if (Inf(dx) == Sup(dx) && Inf(dx) == CXSC_Zero)
      y = adjust(_l_interval(1.0));
   else if (Inf(dx)<-708.396418532264) y = einfachgenau; // Blomquist,12.07.04
   else 
   {
      if (stagprec < stagmax) 
         stagprec++;
      else                    
         stagprec = stagmax;

      if (Sup(dx) <= 0.0)  
         t = -x;  // Werte unter -1e308 nicht auswertbar,
      else                 
         t =  x;  // dafuer andere Ergebnisse exakter
    
      dt = _interval(t);

      // Argumentreduktion
      //    eps = 0.000001*_real(Sup(t))/stagprec;
      //    eps = 1.0/pow10(stagprec/4);
      eps = 0.1;
      while(Sup(dt)/zhn > eps) 
      {
         rednum++;
         zhn += zhn;
      }
      t /= _l_interval(zhn);

      // Taylor Approximation
      dt  = _interval(t);

      // Anzahl der Durchlaeufe bei der Approximation
      tmp = Sup(abs(dt));
      if (MinReal > tmp) 
         tmp = MinReal;
      lny = ln(tmp);
      lneps = (1.0-digits*stagprec)*lnb;
      
      while(lneps-lnb+ln(fak)-_real(n)*lny <= 0.0) 
      {
         n += 3;
         if (n > 170) 
         {    // 170! = 7.26*10306
            cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval exp(const l_interval & x)"));
            n = 170;
            break;
         }
         fak = fak*_real(n)*(n-1.0)*(n-2.0);
       }
      degree = n;

      p = t/_real(degree);
      int i;
      for (i=degree-1; i >= 1; i--)
         p = (p+1.0)*t/_real(i);

      // Fehlerabschaetzung
      error = _interval(-2.0, 2.0)*pow(dt, _interval(_real(n)))/fak;

      // Fehler einbauen
      p += 1.0+_l_interval(error);

      // Argumentreduktion wird rueckgaengig gemacht
      for (i = 1; i <= rednum; i++)
         p *= p;

      if (Sup(dx) < 0.0)  
         p = 1.0/p;

      stagprec = stagsave;
      y = adjust(p);
      y = y & einfachgenau;
   }

   return y;
}

l_interval expmx2(const l_interval& x)
// e^(-x^2);  Blomquist, 13.04.04;
{
    int stagsave = stagprec,
	stagmax = 19;
    l_interval y,z = abs(x);
    l_real r1,r2;
    interval dx = interval(z),
	     einfachgenau;
    einfachgenau = expmx2(dx);
    if (stagprec>stagmax) stagprec = stagmax;

    if (stagprec == 1) y = expmx2(dx); // simple precision 
    else if (Inf(z)>30) y = einfachgenau; // verhindert Overflow bei z*z!
    else y = exp(-z*z); 

    stagprec = stagsave;
    y = adjust(y);
    y = y & einfachgenau;
    return y;
} // expmx2




int cxsc_zweihoch(int) throw();

l_interval ln(const l_interval & x) throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF)    // Ln(x)
{ 
   int         stagsave = stagprec,
               stagmax = 19,
               k,
               m = 0,
               n = 0,
               digits = 53,   // Anzahl der Mantissenstellen
               cmp,
               zhn1 = 1,
               zhn2 = 1;
   bool        zwei=false;
   real        mx, tmp, 
   //            zkp1,
               ln2 = 0.69314718,
               lnb = 0.69314718,
               lny, lneps;
   l_interval  y, t, t2, p;
   interval    dx = _interval(x),
               einfachgenau,
               dy, error;

   einfachgenau = ln(dx);
   // ---------------------------------------------------------------------
   if (Sup(dx) > succ(succ(Inf(dx)))) {  // Dieser Teil vermeidet eine
       y = einfachgenau; y = adjust(y);  // Fehlermeldung der pow(..)-Fkt,
       goto fertig;                      // wenn x zu breit ist und die 1
   }                                     // enthlt. Bei zu breiten Inter-
   // vallen wird y daher mit einfachgenau berechnet.  Blomquist,05.12.03;
   // ---------------------------------------------------------------------
   if (Inf(x) <= 0.0) 
      cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval ln(const l_interval & x)"));
   else if (stagprec == 1) 
      y = ln(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_One)
      y = adjust(_l_interval(0.0));
   else 
   {
      if (stagprec < stagmax) 
         stagprec++;
      else                    
         stagprec = stagmax;

      // erste Argumentreduktion
      /*
         if (Inf(dx) > 1.0) 
         {
            y = 1.0/x;
            dx = _interval(x);
            neg = true;
         } else 
            y = x;
      */
      y = x;

      // zweite Argumentreduktion
      mx = Sup(dx);
      tmp = 1.0/Sup(dx);
      if (tmp > mx) 
         mx=tmp;
      tmp = 1.0/Inf(dx);
      if (tmp > mx) 
         mx = tmp;
      if (mx > 1.1) 
      {
         cmp = int(_double(ln(ln(mx)/ln(1.1))/ln2));
         if (cmp > 0) 
         {
            if (cmp > 14) 
               n = 14;
            else 
               n = cmp;
            zhn1 = cxsc_zweihoch(n);
         }
      }
      y = sqrt(y, zhn1);
      mx = abs(_real(Sup(y)));
      if (mx > 1.1) 
      {
         cmp = int(_double(ln(ln(mx)/ln(1.1))/ln2));
         if (cmp > 0) 
         {
            if (cmp > 14) 
               n = 14;
            else 
               n = cmp;
            zhn2 = cxsc_zweihoch(n);
         }
         y = sqrt(y, zhn2);
         zwei = TRUE;
      }

      // dritte Argumentreduktion
      t = (y-1.0)/(y+1.0);

      // Abschtzung der Genauigkeit
      t2 = t*t;
      tmp = Sup(_interval(t2));
      if (tmp == 0.0) 
         tmp = MinReal;
      dy = tmp;
      lny = Sup(ln(dy));
      lneps = (1.0-digits*stagprec)*lnb;
      
      do {
         m += 2;
      } while (lneps-lnb+ln(2.0*m+3.0)-_real(m+1)*lny <= 0.0);

      // Polynomauswertung
      p = 0.0;
      for (k = 2*m+1; k >= 1; k -= 2)
         p = p*t2+_l_interval(2.0)/_real(k);
      p *= t;

      // Fehlereinschlieung
      error = _interval(-4.0, 4.0)*pow(_interval(t2), _interval(_real(m+1)))/(2.0*m+3.0);

      // Fehler einbauen
      p = p+error;

      // Argumentreduktion 2 wird rckgngig gemacht
      y = _real(zhn1)*p;
      if (zwei) 
         y *= _real(zhn2);

      // Argumentreduktion 1 wird rckgngig gemacht
      /*    if (neg) y = -(y);*/

      stagprec = stagsave;
      y = adjust(y);
      y = y & einfachgenau;

      // Zustzliche Fehlerbehandlung ntig, da Underflow-Fehler nicht abgefangen werden kann
      /*    
         error = _interval(-4.0*pow10(-stagprec*16-1), 4.0*pow10(-stagprec*16-1));
         y += error;
      */

   }
 fertig: // Blomquist,05.12.03;
   return y;
}

int cxsc_zweihoch(int n) throw()
{ 
   // liefert 2^n

   int res = 1;
   switch (n) 
   {
      case 0:
         return 1;
      case 1:
         return 2;
      case 2:
         return 4;
      default:
         {
            int y = 1,
            x = 2,
            zhi = 2;

            if (n%2)  
               res = 2;
            y = x*x;
            do {
               if ((n/zhi)%2)  
                  res *= y;
               zhi += zhi;
               if (zhi <= n)  
                  y *= y;
            } while (zhi <= n);
         }
   }
   return res;
}


l_interval sinh(const l_interval & x) throw(ERROR_LINTERVAL_FAK_OVERFLOW)   // Sinh(x)
{
   long int    n = 1;
   int         stagsave = stagprec,
               stagmax = 19,
               sign = 1,
               digits = 53,
               degree, stagsave2;
   real        tmp,
               lnt, lneps,
               fak = 1.0,
               lnb = 0.69314718;
   l_interval  y,
               t, t2, p, pot;
   interval    dy,
               dx = _interval(x),
               einfachgenau,
               ibase = _interval(2.0),
   //            ilnb = _interval(0.6931471,0.6931472),
   //            ilnx, ilneps, in,
               err;
   //            ifak = _interval(fak);

   einfachgenau = sinh(dx);

   if (stagprec == 1) 
      y = sinh(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
      y = x;
   else 
   {
      if (stagprec < stagmax) 
         stagprec++;
      else                    
         stagprec = stagmax;

      // Ausnuztung der Punktsymmetrie
      if (Sup(dx) < 0.0) 
      {
         y = -x;
         sign = -1;
      } else  
         y =  x;
    
      dy = _interval(y);

      // Bei Werten ber 0.5 -- Auswertung ber e-Funktion
      if (Sup(dy) > 0.5) 
      {
         try {
            t = exp(y);
         }
         catch(const ERROR_LINTERVAL_FAK_OVERFLOW &)
         {
            cxscthrow( ERROR_LINTERVAL_FAK_OVERFLOW("l_interval sinh(const l_interval & x)")); 
         }
         y = sign*0.5*(t-1.0/t);
      }
      // Auswertung ber Potenzreihenentwicklung
      else 
      {
         t = y;
         t2 = t*t;

         // Abschtzung der Genauigkeit
         tmp = _real(Sup(abs(t)));       // bei Andreas und Christian "t2"
         if (tmp < MinReal)  
            tmp = MinReal;
         lnt = ln(tmp);
         lneps = (1.0-digits*stagprec)*lnb;
         do {
            n += 3;
            if (n > 170) 
            {    // 170! = 7.26*10^306
               cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval sinh(const l_interval & x)"));
               n = 170;
               fak *= n*(n-1.0);
               break;
            }
            fak *= n*(n-1.0)*(n-2.0);
         } while(lneps+ln(fak)-_real(n)*lnt-lnb <= 0.0);
         /*
            real bas, tn, t2d;
            bas = _double(_real(power(_l_real(2.0),1-digits*stagprec)));
            tn = abs(_double(_real(mid(t))));
            t2d = _double(_real(mid(t2)));
            do {
               n += 2.0;
               if (n > 170) 
               {    // 170! = 7.26*10^306
                  errmon(ERROR_LINTERVAL(FAKOVERFLOW));
                  n = 170;
                  fak *= n;
                  break;
               }
               tn *= t2d;
               fak *= n*(n-1.0);
            } while(bas-tn/fak <= 0.0);
         */
         degree = 2*(n/2+1);  // degree ist gerade
                              // Achtung degree = 2n+2!

         // Auswertung mit Horner-Schema
         p = 1.0;
         for (int i = degree; i >= 2; i -= 2) 
         {
            pot = _interval((i+1.0)*i);
            p = (p*(t2))/pot+1.0;
         }
         p *= t;

         // Negative Werte zurckwandeln
         if (sign == -1)  
            p = -p;

         // Fehlerauswertung
         stagsave2 = stagprec;
         stagprec = 2;
         l_interval  relerr;
         stagprec = stagsave2;
         err = pow(ibase , _interval(1.0-digits*stagprec)) * _interval(-1.0, 1.0);
         relerr[1] = 1.0;
         relerr[2] = Inf(err);
         relerr[3] = Sup(err);

         // Fehler einbauen
         y = p*relerr;
      }
      stagprec = stagsave;
      y = adjust(y);
      y = y & einfachgenau;
   }
   return y;
}

l_interval cosh(const l_interval & x) throw(ERROR_LINTERVAL_FAK_OVERFLOW)   // Cosh(x)
{
   int         stagsave = stagprec,
               stagmax = 19;
   l_interval  y, t;
   interval    dx = _interval(x),
               einfachgenau;

   einfachgenau = cosh(dx);

   if (stagprec == 1) 
      y = cosh(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
      y = adjust(_l_interval(1.0));
   else 
   {
      if (stagprec < stagmax) 
         stagprec++;
      else                    
         stagprec = stagmax;
      if (Sup(dx) <= 0.0)  
         y = -x;
      else                 
         y =  x;
      
      try 
      {
         t = exp(y);
      } 
      catch(const  ERROR_LINTERVAL_FAK_OVERFLOW &)
      {
         cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval cosh(const l_interval & x)"));
      }
      
      y = 0.5*(t+1.0/t);
      if (_interval(0.0) <= dx) 
         SetInf(y,1.0);
      stagprec = stagsave;
      y = adjust(y);
      y = y & einfachgenau;
   }

   return y;
}

// Fields for tanh-function
int  ex_tanh[8] = { 1,-1,-2,-4,-5,-6,-8,-9 };  
int  c_tanhN[8] = { 1,-1,2,-17,62,-1382,21844,-929569 };
int  c_tanhD[8] = { 1,3,15,315,2835,155925,6081075,638512875 }; 

l_interval tanh(const l_interval & x) throw()
// tanh(x) with Taylor-series and x --> MaxReal; Blomquist 24.05.04;           
{
    l_interval s,t,y;
    interval dx = _interval(x),
             einfachgenau,r,r2;
    int stagsave = stagprec,ex,m,N,k,
	stagmax = 19;
    bool neg;
    real Sdx;
    einfachgenau = tanh(dx); 
    if (stagprec>stagmax) stagprec = stagmax;
    if (stagprec == 1) y = tanh(dx);
    else if (dx==0) y = 0;
    else if (0<=dx) y = einfachgenau;
    else // 0 not in x and not in dx:
    {   
	t = x;  
	neg = Sup(dx)<0;  Sdx = Sup(dx);
	if (neg) {t = -t; dx = -dx; Sdx = Sup(dx);} // Inf(dx) > 0:
	ex = expo(Sdx);
	if (ex < -70) {
	    m = ex - 53*stagprec; // m >= -1074 necessary !!
	    if (m < -1074) m = -1074;
	    r = interval(Sdx); 
	    r2 = r*r;
	    N = 0;
	    do {
		N++;
		r = r*r2;
		k = expo(Sup(r)) + ex_tanh[N];
	    } while (k>m);
	    r2 = interval(c_tanhN[N])/c_tanhD[N];
	    r2 = r * abs(r2); // r2: inclusion of the absolute error
	    N--;  // N: Polynomial degree
	    y = l_interval(c_tanhN[N])/c_tanhD[N];
	    s = t*t;
	    for (k=N-1; k>=0; k--)  // Horner-scheme
		y = y*s + l_interval(c_tanhN[k])/c_tanhD[k];
	    y = y * t;
	    y = y + interval(-Sup(r2),Sup(r2)); // Approxim. error
	} else if (ex<-4) {
	    s = sinh(t);  y = cosh(t);
	    if (stagprec<stagmax) stagprec++;
	    y = s/y;
	} else if (Sdx<352) {
	    times2pown(t,1);
	    y = 1 - 2/(exp(t)+1); // --> Sup(y) <= 1 !!
	} else {
	    if (Inf(dx)<352) {
		t = Inf(t);  times2pown(t,1);
		y = 1 - 2/(exp(t)+1);
	    } else y = l_interval(1) - comp(0.5,-1013);
	    SetSup(y,1); 
	}
	if (neg) y = -y;
    } // else

    stagprec = stagsave; // restore old stagprec value
    y = adjust(y);  // adjust precision of y to old stagprec value
    y = y & einfachgenau; // optimal inclusion in case of too
                          // large relative diameters of y;
    return y;
} // tanh

// ************************************************************************
// Fields for coth-function
int  ex_coth[8] = { -1,-5,-8,-12,-15,-18,-22,-25 };  
int  c_cothN[8] = { 1,-1,2,-1,2,-1382,4,-3617 };
real c_cothD[8] = { 1,15,315,1575,31185,212837625,
                    6081075,54273594375.0 }; // components exactly stored!
// ************************************************************************

l_interval coth(const l_interval & x) throw()
// coth(x); Blomquist 17.04.04;           
{
    l_interval t,s,c,y;
    interval dx = _interval(x),
             einfachgenau,r,r2;
    int stagsave = stagprec,ex,m,N,k,
	stagmax = 19;
    bool neg;
    einfachgenau = coth(dx); // produces all necessary error messages!

    if (stagprec == 1) y = coth(dx);
    else 
    {
	if (stagprec>stagmax) stagprec = stagmax;
	neg = Sup(dx)<0.0;
	t = x;
	if (neg) { t = -t; dx = -dx; } // Inf(t),Inf(dx) > 0;
	ex = expo(Inf(dx));
	if (ex<-66) { // Laurent series
	    m = -ex - 53*stagprec;
	    r = interval(Sup(dx)); r2 = r*r;
	    N = 0;
	    do { 
		N++;
		if (N>7) { y = einfachgenau; goto Fertig; }
		r = r*r2;
		k = expo(Sup(r)) + ex_coth[N];
	    } while (k>m);
	    r2 = interval(c_cothN[N])/c_cothD[N]; 
	    r2 = r*abs(r2)/3;  // r2: inclusion of the absolute error
	    N--; // Polynomial degree
	    y = l_interval(c_cothN[N])/c_cothD[N]; 
	    s = t*t;
	    for (k=N-1; k>=0; k--)
		y = y*s + l_interval(c_cothN[k])/c_cothD[k];
	    y = (y*t)/3 + 1/t;
	    y = y + interval(-Sup(r2),Sup(r2)); // with approx. error
	} else if (ex < 2) {
	    if (stagprec<stagmax) stagprec++; // stagprec <= 19
	    c = cosh(t);
	    s = sinh(t);
	    y = c/s;
	} else if (Inf(dx)<353.0) {
	    if (stagprec<stagmax) stagprec++; // stagprec <= 19
	    times2pown(t,1);
	    y = 1.0 + 2.0/(exp(t)-1.0);
	} else { // Inf(dx) >= 353.0
	    y = l_interval(1.0) + comp(0.5,-1016);
	    SetInf(y,1.0);
	}
	if (_interval(1.0) <= y)  SetInf(y,1.0);
	if (neg) y = -y;
    Fertig:	stagprec = stagsave; // restore old stagprec value
	y = adjust(y);  // adjust precision of y to old stagprec value
	y = y & einfachgenau;
    }
    return y;
} // coth


l_interval acosh(const l_interval & x) throw()  
// acosh(x); Blomquist 14.04.04;
{
    int         stagsave = stagprec,ex1,ex2,
                stagmax = 19;
    interval    dx = interval(x),
                einfachgenau;
    l_interval  y,t;

    einfachgenau = acosh(dx); // false definition range stops program
      
    if (stagprec == 1) y = acosh(dx);
    else if (Inf(dx) == Sup(dx) && Sup(dx) == 1.0)
	y = adjust(_l_interval(0.0));
    else  
    {
	if (stagprec < stagmax) 
            stagprec++;
	else stagprec = stagmax;
	ex1 = expo(Inf(dx));  ex2 = expo(Sup(dx));
	if (ex1>500) { // acosh(x) approx ln(2) + ln(x):
	    y = li_ln2() + ln(x);
	    // Absolute approximation error:
	    t = 1.0/Inf(x);
	    y += l_interval(-Sup(t*t),0); // approxim. error realized here
	} else if (ex2<2) {
	    t = x-1;
	    y = lnp1( t+sqrt(t*(2+t)) );
	} else y = ln(x+sqrtx2m1(x));

	stagprec = stagsave;
	y = adjust(y);
	y = y & einfachgenau;
    } 
   return y;
}

l_interval asinh(const l_interval & x) 
    throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF,ERROR_LINTERVAL_FAK_OVERFLOW) 
// ASinh(x) hier ohne zeitaufwendiges Newton-Verfahren, da die Funktion 
// sqrt1px2(x) = sqrt(1+x*x) zur Verfuegung steht, die Overflow vermeidet!
// Blomquist, 28.12.03;
{
   int         stagsave = stagprec,
               stagmax = 19;
   l_interval  y, my;
   interval    dx = _interval(x), einfachgenau, 
               error = abs(dx);
   real absSupdx = Sup(error), r;

   try 
   {
      einfachgenau = asinh(dx);

      if (stagprec == 1) 
            y = asinh(dx);
      else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
            y = x;
      else // if (absSupdx < MaxReal) 
      {
         if (stagprec < stagmax) stagprec++;
         else stagprec = stagmax; // Erhoete Praezision:

	 if (absSupdx < 2E-108) { // Taylorreihe von ln(x+sqrt(1+x2))
         // ln(x+sqrt(1+x2)) = x - x3/6 +- .....
	     y = x;
	     dx = interval(absSupdx);
	     error = dx*dx*dx/6;
	     r = Sup(error);  // r: Oberschranke des absoluten Fehlers
         // Jetzt folgt die Addition einer garantierten Einschliessung 
         // des abs. Fehlers:
	     y += l_interval(-r,r); 
	 } else
	     if (Sup(x) < 0.0) 
		 if (absSupdx < 1E+10) y = lnp1( -x+sqrtp1m1(x*x) );  
		 else y = -ln(-x+sqrt1px2(x));
	     else 
		 if (absSupdx < 1E+10) y = lnp1( x+sqrtp1m1(x*x) );
		 else y = ln(x+sqrt1px2(x));

         stagprec = stagsave;
         y = adjust(y);
	 y = y & einfachgenau;
      } 
   } // try 
   catch(const ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF &)
   {
      cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval asinh(const l_interval & x)"));
   }
   catch(const ERROR_LINTERVAL_FAK_OVERFLOW &)
   {
      cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval asinh(const l_interval & x)"));
   }
   return y;
} // asinh(x)

l_interval atanh(const l_interval & x) 
   throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF,ERROR_LINTERVAL_FAK_OVERFLOW) 
// ATanh(x), Blomquist 29.12.03;
{
   int         stagsave = stagprec,
               stagmax = 19;
   l_interval  y, my;
   interval    dx = _interval(x), 
               error = abs(dx),
               einfachgenau;
   real absSupdx = Sup(error);

   einfachgenau = atanh(dx);

   if (Inf(x) <= CXSC_MinusOne || Sup(x) >= CXSC_One) 
      cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval atanh(const l_interval & x)"));
   else if (stagprec == 1)
      y = atanh(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
      y = x;
   else 
   {
      if (stagprec < stagmax) stagprec++;
      else stagprec = stagmax; // Ab jetzt hoehere Praezision
      if (absSupdx < 0.125) {
	  y = x / (1-x);
	  times2pown(y,1);  // Schnelle Multiplikation mit 2
	  y = lnp1(y);
	  times2pown(y,-1); // Schnelle Division durch 2
      } else { 
	  y = ln((1.0+x)/(1.0-x));
	  times2pown(y,-1);
      }
      stagprec = stagsave;
      y = adjust(y);
      y = y & einfachgenau;
   } 

   return y;
} // atanh

l_interval acoth(const l_interval & x) 
    throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF,ERROR_LINTERVAL_FAK_OVERFLOW)
// Acoth(x) hier ohne das langsame Newtonverfahren; Fue grosse x gilt:
// Acoth = 0.5*ln(1+2/(x-1));  Blomquist, 28.12.03;
{
   int         stagsave = stagprec,
               stagmax = 19;
   l_interval  y, my;
   interval    dx = _interval(x),
               einfachgenau,
               error = abs(dx);
   real absSupdx = Sup(error);

   einfachgenau = acoth(dx);

   if ((_l_interval(Inf(x))) < _l_interval(CXSC_MinusOne,CXSC_One)
    || (_l_interval(Sup(x))) < _l_interval(CXSC_MinusOne,CXSC_One))
      cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF
                       ("l_interval acoth(const l_interval & x)"));
   else if (stagprec == 1)
      y = acoth(dx);
   else  
   {
      if (stagprec < stagmax) stagprec++;
      else stagprec = stagmax; // Rechg. in erhoeter Praezision:

      if (absSupdx > 1E10) {
	  y = l_interval(2)/(x-1);
	  y = lnp1(y);
	  times2pown(y,-1);
      } else { 
	  y = ln((x+1.0)/(x-1.0));
	  times2pown(y,-1);
      }

      stagprec = stagsave;
      y = adjust(y);
      y = y & einfachgenau;
   } 

   return y;
} // acoth

l_interval lnp1(const l_interval& x) throw()
// ln(1+x) = zeta * P(zeta);  zeta := x/(2+x);
// P(zeta) := 2 + (2/3)zeta^2 + (2/5)zeta^4 + ...
// Approximation of P(zeta) with P_N(zeta), defined by
// P_N(zeta) := 2 + (2/3)zeta^(2*1) + 
//                + (2/5)zeta^(2*2) + ... + (2/2N+1)zeta^(2*N);
// Absolute approximation error Delta := P(zeta) - P_N(zeta);
// With z:=zeta^2  it holds:    Delta <= (2*z^(N+1))/{(2N+3)*(1-z)};
{
    const real c1(0.693147181);
    int stagsave(stagprec),
	stagmax(19);  // For ln-function
    if (stagprec>stagmax) stagprec = stagmax; // stagprec now <= 19;
    interval dx(x),    // dx of type interval inclusing x.
	einfachgenau(lnp1(dx)), // einfachgenau inclusing lnp1(x).
	ax( abs(dx) ); // ax inclusing the absolute values of dx.
    real gr(Sup(ax)),  // gr: maximum of the absolute values from x.
	lngr,u;
    l_interval t(x);
    if (gr < 1E-8)  // Series expansion, gr = 1E-8 is Ok!
	if (sign(gr)==0) t=0;
	else { // Calculation of N for the approximation of P(zeta)
               // using the polynomial P_N(zeta) of degree N: 
	    int N(0),TwoNp3(3),k;
	    k = expo(gr);
	    if (k>-1019) {
	    k = k - 1 - 53*stagprec;
	    lngr = ln(gr/2);
	    if (k < -1074) k = -1074;
	    k--; 
	    u = k*c1;  // u = (k-1)*ln(2);
	    do {
		N++;
		TwoNp3 += 2;
	    } while(TwoNp3*lngr - ln(TwoNp3) > u);
	    }  	
            // Polynomial degree 0<= N <=17 is now calculated!
            // Calculation of the N+1 polynomial coefficients:
	    l_interval a_lnp1[18], // Declaration of the coeffits.
		zeta,z2;
	    a_lnp1[0] = 2;
	    for (int i=1; i<=N; ++i) {
		a_lnp1[i] = a_lnp1[0]/(2*i + 1); 
	    }
            // Polyn.-coeff. a_lnp1[i] are now calculated; 0<=i<=N;
            // The calculation of P_N(zeta) follows now:
	    zeta = t/(2+t);
	    z2 = sqr(zeta);  
	    dx = z2; // is used for the approximation error!
	    t = a_lnp1[N]; // Horner-scheme:
	    for (int i=N-1; i>=0; --i) {
		t = t*z2 + a_lnp1[i];
	    } // t = P_N(zeta) is now calculated
            // Calculating now the approximation error 
            // with dx of type interval:
	    dx = interval(Sup(dx));
	    ax = Power(dx,N+1); 
	    times2pown(ax,1); // Multiplication with 2;
	    dx = ax / ((2*N+3)*(1-dx));
	    t = t + l_interval(0.0,Sup(dx)); 
            // Approximation error implemented now 
	    t = zeta * t;
	}
    else 
	if (gr<1) { // Calculation in higher accuracy:
	    stagprec++;
	    if (stagprec>stagmax) stagprec = stagmax;
	    t = ln(1+t);
	} else t = ln(1+t); // Normal accuracy:
    stagprec = stagsave;
    t = adjust(t);
    t = t & einfachgenau; // intersection of t and einfachgenau;
    // If x is too wide and therefore t because of the inevitable 
    // interval overestimations, t is the best possible inclusion.  
    return t;
}

l_interval sqrtx2m1(const l_interval& x)
// sqrt(x^2-1);  Blomquist, 13.04.04;
{
    int stagsave = stagprec,
	stagmax = 19;
    l_interval y,z = abs(x);
    l_real r1,r2;
    interval dx = _interval(z),
	     einfachgenau;
    einfachgenau = sqrtx2m1(dx);
    if (stagprec>stagmax) stagprec = stagmax;

    if (stagprec == 1) y = sqrtx2m1(dx); // simple interval 
    else if (Inf(z)==1) {
	r1=0;  z = Sup(z);  z = sqrt(z*z-1);
	r2 = Sup(z);
	y = l_interval(r1,r2);
    } else if (expo(Sup(dx))<500) y = sqrt(z*z-1.0); // no overflow!
    else { // avoiding overflow using:  x-1/(2x) < sqrt(x^2-1) < x
	r2 = Sup(z);
	z = Inf(z);  // z: Point interval
	y = 1.0/z;
	times2pown(y,-1);
	y = z - y;
	r1 = Inf(y);
	y = l_interval(r1,r2);
    } 
    stagprec = stagsave;
    y = adjust(y);
    y = y & einfachgenau;
    return y;
} // sqrtx2m1

l_interval sqrt1mx2(const l_interval& x)
// sqrt(1-x^2);  Blomquist, 13.04.04;
{
    int stagsave = stagprec,
	stagmax = 19;
    l_interval y,z = abs(x);
    l_real r1,r2;
    interval dx = interval(z),
	     einfachgenau;
    einfachgenau = sqrt1mx2(dx);
    if (stagprec>stagmax) stagprec = stagmax;

    if (stagprec == 1) y = sqrt1mx2(dx); // simple interval 
    else {
	y = comp(0.5,1023); // y = 2^(+1022)
	times2pown(z,511);  
	y = sqrt(y-z*z);
	times2pown(y,-511);
    } 
    stagprec = stagsave;
    y = adjust(y);
    y = y & einfachgenau;
    return y;
} // sqrt1mx2


l_interval ln_sqrtx2y2(const l_interval& x, const l_interval& y) throw()
{ // Inclusion of ln(sqrt{x^2+y^2}) in staggered arithmetic
    int stagsave = stagprec;
//	stagmax = 19;  
    interval dx = interval(x),
	dy = interval(y),
	einfachgenau = ln_sqrtx2y2(dx,dy);
    dx = abs(dx);  dy = abs(dy);
    l_interval ax(abs(x)),ay(abs(y)),ar; 
    int ex_x(expo(Sup(dx))), ex_y(expo(Sup(dy))),
	N,N1;
    if (ex_y>ex_x) ex_x = ex_y;
    if (ex_x>508) { // To avoid overflow:
	N = ex_x-500;
	times2pown(ax,-N); times2pown(ay,-N);
	ar = ax*ax + ay*ay;
	ar = ln(ar);
	times2pown(ar,-1);
	ar += N*li_ln2();
    } else 
	if (ex_x<-20) { // To avoid underflow:
	    N = 500 - ex_x;
	    if (N>1023) { // Avoiding an range error with times2pown(...)  
		N1 = N-1023;
		times2pown(ax,1023);  times2pown(ax,N1); 
		times2pown(ay,1023);  times2pown(ay,N1);
	    } else {
		times2pown(ax,N); 
		times2pown(ay,N);
	    }
	    ar = ax*ax + ay*ay;
	    ar = ln(ar);
	    times2pown(ar,-1);
	    ar -= N*li_ln2();  // ar = ar - N*ln(2)
	} else { // Calculation with function lnp1:
	    ar = sqr(ax) + sqr(ay) - 1;
	    ar = lnp1(ar);
	    times2pown(ar,-1); 
	}
    stagprec = stagsave;
    ar = adjust(ar);
    ar = ar & einfachgenau;

    return ar;
}

l_interval acoshp1(const l_interval& x)
// acoshp1(x) = acosh(1+x);  Blomquist, 20.04.05;
{
    int stagsave = stagprec,ex,
	stagmax = 19;
    l_interval y,t;
    l_real lr;
    interval dx = interval(x),
	     einfachgenau;
    einfachgenau = acoshp1(dx);
    if (stagprec>stagmax) stagprec = stagmax;

    if (stagprec == 1) y = acoshp1(dx); // simple interval 
    else {
	ex = expo(Sup(dx));
	if (ex<=-1016) {
	    t = x;
	    times2pown(t,1);  // fast multiplication with 2 = 2^1;
	    t = sqrt(t);      // t = sqrt(2x);
	    lr = Inf(t);      // Calculating the infimum (Leibniz-series)
	    y = l_interval(lr)*(1.0 - l_interval(lr)/12.0);
	    y = l_interval(Inf(y),Sup(t)); // using alternating Leibniz-series
	}
	else 
	    if (ex<-400) y = lnp1(x*(1.0+sqrt(1+2.0/x)));
	    else y = acosh(1+x); // x = MaxReal not allowed
    } 
    stagprec = stagsave;
    y = adjust(y);
    y = y & einfachgenau;
    return y;
} // acoshp1


} // namespace cxsc
