/*
**  CXSC is a C++ library for eXtended Scientific Computing (V 2.1.1)
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2006 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* CVS $Id: l_rmath.cpp,v 1.11 2006/03/08 14:16:27 cxsc Exp $ */

#include "l_rmath.hpp"
#include "rmath.hpp"

namespace cxsc {

l_real sqrt(const l_real &x) throw(ERROR_LREAL_STD_FKT_OUT_OF_DEF)
// Blomquist, additional scaling, 10.12.02
{
   int stagsave = stagprec, stagmax = 19, stagcalc;
   l_real y;
   if (sign(x[1])<0) // Blomquist: is faster than  x < 0.0;
   cxscthrow(ERROR_LREAL_STD_FKT_OUT_OF_DEF("l_real sqrt(const l_real &x)"));
   else if (zero_(x) || x == 1.0 ) // Blomquist, faster than x == 0;
      y = x;
   else 
   {
       l_real x1 = x;
       int ex = expo(x1[1]);
       ex = 1021-ex;        // ex: optimal scaling factor
       if (ex%2) ex--;      // ex is now even;
       times2pown(x1,ex);   // scaling x1 with 2^ex

       // Einsatz des Newton-Verfahrens y = y-f(y)/f'(y)
       if (stagprec < stagmax) stagcalc = stagprec+1;  
       else stagcalc = stagmax+1;  // Blomquist with +1; 30.11.02;
       y = sqrt(_real(x1));
       stagprec = 1; 
       while (stagprec < stagcalc) 
       {
	   stagprec += stagprec; // now calculation in double precision:
           if (stagprec > stagcalc) stagprec = stagcalc;
           // y = 0.5*((x1/y)+y);
	   y += x1/y;  // Blomquist, 30.11.02;
	   times2pown(y,-1); // Blomquist, this multiplication is faster!!
       }
       times2pown(y,-ex/2); // backscaling
       stagprec = stagsave; // restore old stagprec
       adjust(y);
   }
   return y;
}

l_real power(const l_real& x, int n)
{
   int       stagsave = stagprec, stagmax = 19;
   long int  zhi = 2;
   l_real    y, neu;
   bool      neg=false;

   if (x == 1.0)
      y = x;
   else if (n == 0)
      y = adjust(_l_real(1.0));
   else 
   {
      if (stagprec < stagmax) 
         stagprec++;
      else
         stagprec = stagmax;

      if (n == 1)
         y = x;
      else if (n == 2)
         y = x*x;
      else 
      {
         if (n < 0) 
         {
            neg = true;
            n = -n;
         }

         // Initialisierung
         if (n%2)   
            y = x;
         else      
            y = 1.0;  // Praezision wird bei 1 Mult. auf
                      // aktuellen Wert gesetzt;
                      // Berechnung durch binaere Darstellung der n
         neu = x*x;
         do {
            if ((n/zhi)%2)
               y *= neu;
            zhi += zhi;
            if (zhi <= n)  // letzte Mult. entfaellt --> schneller
               neu *= neu;
         } while (zhi <= n);
         if (neg) 
            y = 1/(y);
      }
      stagprec = stagsave;
      y = adjust(y);
   }
   return y;
}

} // namespace cxsc
